/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include "Editors/Editor.h"
#include "Midi/Track.h"
#include "Midi/Sequence.h"
#include "GUI/GraphicalTrack.h"
#include "Pickers/MagneticGrid.h"
#include "Pickers/VolumeSlider.h"
#include "Config.h"
#include "Images/ImageProvider.h"
#include "Images/Drawable.h"
#include "GUI/GLPane.h"
#include "Pickers/InstrumentChoice.h"

#include "OpenGL.h"

namespace AriaMaestosa {
	
Editor::Editor(Track* track)
{
	INIT_LEAK_CHECK();
	
    Editor::track = track;
    Editor::sequence = track->sequence;
    Editor::graphicalTrack = track->graphics;
    
    volumeSlider=new VolumeSlider();
    
    verticalScrolling=false;
    click_on_scrollbar=false;
	
    scroll_up_arrow_pressed=false;
    scroll_down_arrow_pressed=false;
	
	selecting = false;
	
	instrument=new InstrumentChoice(graphicalTrack);
}

Editor::~Editor()
{
    delete volumeSlider;
	delete instrument;
}

void Editor::render(const int mousex_current, const int mousey_current, const int mousex_initial, const int mousey_initial, const bool focus)
{
}

void Editor::updatePosition(int from_y, int to_y, int width, int height, int barHeight)
{
    Editor::from_y = from_y;
    Editor::to_y = to_y;
    Editor::width = width;
    Editor::height = height;
    Editor::barHeight = barHeight;
}

void Editor::rightClick(int x, int y)
{
    std::cout << "Edit base class called. This should not happen." << std::endl;
    exit(0);

}

void Editor::drawVerticalMeasureLines(const int from_y, const int to_y)
{
    
    glLoadIdentity();
    glDisable(GL_TEXTURE_2D);
    glLineWidth(1);
    int count=1;
    
    const float zoom=sequence->getZoom();
    const float increment=(float)sequence->ticksPerMeasure()/(float)sequence->getMeasureTopValue()*zoom;
	
	if(increment == 0)
	{
		std::cout << "Fatal Error: increment=0 in Editor::drawVerticalMeasureLines (" << sequence->ticksPerMeasure() << " / " << sequence->getMeasureTopValue() << " * " << zoom  << ")"<< std::endl;
		wxMessageBox( wxT("Fatal Error: increment=0 in Editor::drawVerticalMeasureLines"));
		exit(1);
	}
	
    const int end = getXEnd();
    /*
    std::cout << "zoom: " << zoom << "\ngetEditorXStart(): " << getEditorXStart() << "\nsequence->ticksPerMeasure(): " << sequence->ticksPerMeasure() <<
        "\nsequence->getMeasureTopValue(): " << sequence->getMeasureTopValue() << "\nsequence->getXScrollInPixels(): " << sequence->getXScrollInPixels() <<
        "\nincrement: " << increment <<
        "\nmx: " << getEditorXStart()+sequence->ticksPerMeasure()/sequence->getMeasureTopValue()*zoom -
        ((int)(sequence->getXScrollInPixels()/zoom) % sequence->ticksPerMeasure())*zoom << std::endl;
    */
    for(float mx=getEditorXStart()+sequence->ticksPerMeasure()/sequence->getMeasureTopValue()*zoom - ((int)(sequence->getXScrollInPixels()/zoom) % sequence->ticksPerMeasure())*zoom;
        mx<end;
        mx+=increment)
	{
        
        if(count % sequence->getMeasureTopValue() == 0) glColor3f(0.5, 0.5, 0.5);
        else glColor3f(0.9, 0.9, 0.9);
        count++;
        
        if(mx<getEditorXStart()) continue; // don't draw lines before the beginning
		
		glBegin(GL_LINES);
		glVertex2f(mx, from_y);
		glVertex2f(mx, to_y);
		glEnd();
		
        
    }
}

void Editor::renderScrollbar()
{
    
	// ---- scrollbar background -----
    sbBackgDrawable->move(getWidth() - 24, from_y+20+barHeight);
    sbBackgDrawable->scale(1, height/sbBackgDrawable->image->height);
    sbBackgDrawable->render();
    
	// ------ top arrow ------
    sbArrowDrawable->setFlip(false, false);
    sbArrowDrawable->move(getWidth() - 24, from_y+20+barHeight);
    
    const int mouseX = sequence->glPane->getMouseX_initial();
    const int mouseX2 = sequence->glPane->getMouseX_current();
    
    const int mouseY = sequence->glPane->getMouseY_initial();
    const int mouseY2 = sequence->glPane->getMouseY_current();
    
    if(!verticalScrolling and // ignore arrows if user is dragging thumb
	   sequence->glPane->isMouseDown() and // only if mouse is down
       mouseX > sbArrowDrawable->x and mouseX2 > sbArrowDrawable->x and // and mouse is located on the arrow
       mouseY > sbArrowDrawable->y and mouseY2 > sbArrowDrawable->y and
       mouseX < sbArrowDrawable->x+sbArrowDrawable->image->width and mouseX2 < sbArrowDrawable->x+sbArrowDrawable->image->width and
       mouseY < sbArrowDrawable->y+sbArrowDrawable->image->height and mouseY2 < sbArrowDrawable->y+sbArrowDrawable->image->height)
    {
        
        sbArrowDrawable->setImage( sbArrowDownImg );
        scroll_up_arrow_pressed=true;
        
    }
	else
	{
        
        sbArrowDrawable->setImage( sbArrowImg );
        scroll_up_arrow_pressed=false;
        
    }
    
    sbArrowDrawable->render();
    
	// ------ bottom arrow ------
    sbArrowDrawable->setFlip(false, true);
    sbArrowDrawable->move(getWidth() - 24, from_y+barHeight+height+12);
    
    if(!verticalScrolling and // ignore arrows if user is dragging thumb
	   sequence->glPane->isMouseDown() and // only if mouse is down
       mouseX > sbArrowDrawable->x and mouseX2 > sbArrowDrawable->x and // and mouse is lcoated on the arrow
       mouseY > sbArrowDrawable->y and mouseY2 > sbArrowDrawable->y and
       mouseX < sbArrowDrawable->x+sbArrowDrawable->image->width and mouseX2 < sbArrowDrawable->x+sbArrowDrawable->image->width and
       mouseY < sbArrowDrawable->y+sbArrowDrawable->image->height and mouseY2 < sbArrowDrawable->y+sbArrowDrawable->image->height)
    {
        
        sbArrowDrawable->setImage( sbArrowDownImg );
        scroll_down_arrow_pressed=true;
    }
	else
	{
        
        sbArrowDrawable->setImage( sbArrowImg );
        scroll_down_arrow_pressed=false;
    }
    
    sbArrowDrawable->render();
    
	// -------- thumb ------
    sbBarDrawable->move(getWidth() - 24, (int)(from_y+barHeight+37+(height-36)*sb_position));
    sbBarDrawable->render();
    
}


void Editor::mouseDown(int x, int y)
{
    

    lastDragY=y;
    
	click_on_scrollbar = false;
	
    // check if user is grabbing the scroll bar
    if(x>sbBarDrawable->x and
       y<sbBarDrawable->x+sbBarDrawable->image->width and
	   y<getYEnd()-15 and y>getEditorYStart())
	{
        
		click_on_scrollbar = true;
		
        // grabbing the thumb
        const int thumb_pos = (int)(from_y+barHeight+37+(height-36)*sb_position);
        if(y>thumb_pos and y<thumb_pos + sbBarDrawable->image->height)
		{
            verticalScrolling = true; 
        }
        
    }//endif
    
}

void Editor::mouseUp(int mousex_current, int mousey_current, int mousex_initial, int mousey_initial)
{
  verticalScrolling=false;
  click_on_scrollbar=false;
}

void Editor::mouseExited(int dragX_arg, int mousey_current, int XBeforeDrag_arg, int mousey_initial)
{
	
}

/*
 * Is it necessary to send frequent mouse held down events?
 */
bool Editor::areMouseHeldDownEventsNeeded(){

	// if user is clicking on scrollbar but not on thumb, it is needed
	if(click_on_scrollbar and
	   not scroll_up_arrow_pressed and
	   not scroll_down_arrow_pressed and
	   not verticalScrolling
	   )
	{
		return true;
	}
	
	// is user is clicking on a scroll arrow, it is necessary
	if(scroll_up_arrow_pressed or scroll_down_arrow_pressed) return true;
	
	// otherwise, it is not necessary
	return false;  
}

void Editor::mouseHeldDown()
{
	
	click_on_scrollbar = false;
	
	 const int mousex_current = sequence->glPane->getMouseX_current();
	 const int mousey_current = sequence->glPane->getMouseY_current();
	 
    // -------------- check if user is clicking on scroll bar, but not grabbing the thumb ----------------
    // make thumb slide to mouse
    if(mousex_current>sbBarDrawable->x and
       mousey_current<sbBarDrawable->x+sbBarDrawable->image->width and
	   mousey_current<getYEnd()-15 and mousey_current>getEditorYStart())
	{
        
		click_on_scrollbar = true;
		
        // not grabbing the thumb
        if(
           not scroll_up_arrow_pressed and
           not scroll_down_arrow_pressed and
           not verticalScrolling
           )
		{
            
			// find where mouse is on scrollbar, in the same units at sb_position (0 to 1)
            const float goal = (mousey_current-from_y-barHeight-52)/(float)(height-36);
            
			// thumb has reached mouse, just continue as regular scrolling
			float dist = goal-sb_position; // find distance betweem mouse and thumb in 0-1 units
			if(dist<0) dist -= dist*2; // make positive
			if(dist<0.01){
				verticalScrolling = true;
			}
			// move thumb towards mouse
            if(sb_position < goal) sb_position += 0.005;
            if(sb_position > goal) sb_position -= 0.005;
			
			// make sure thumb stays within bounds
			if(sb_position<0) sb_position=0;
			if(sb_position>1) sb_position=1;
			
            sequence->glPane->render();
        }
            
    }//endif
        
    // -------------- check if user is clicking on scroll bar arrows ----------------
    if(scroll_up_arrow_pressed)
	{
        sb_position -= 0.001;
		if(sb_position<0) sb_position=0;

        sequence->glPane->render();
    }
	
    if(scroll_down_arrow_pressed)
	{
        sb_position += 0.001;
		if(sb_position>1) sb_position=1;
		
        sequence->glPane->render();
    }
    
}

void Editor::mouseDrag(int mousex_current, int mousey_current, int mousex_initial, int mousey_initial)
{

    // -------------------  drag scrollbar  ---------------
    if(verticalScrolling and /*mousey_current*/mousey_initial<from_y+barHeight+height+12 and /*mousey_current*/mousey_initial>from_y+20+barHeight)
	{
        
		// if we reach scrollbar extremity, wait until mouse comes back to the middle of the thumb before starting to scroll again
		if( sb_position == 0 and mousey_current<from_y+47+barHeight)
		{
			lastDragY = mousey_current;
			return; 
		}
		if( sb_position == 1 and mousey_current>from_y+barHeight+height+10)
		{
			lastDragY = mousey_current;
			return; 
		}
		
        sb_position += (mousey_current - lastDragY)/(float)(height-36);
        
        if(sb_position<0) sb_position=0;
        if(sb_position>1) sb_position=1;
        
    }
    
    lastDragY = mousey_current;
    
}

const int getEditorXStart()	{	return 90;								}
const int Editor::getXEnd()			{	return width - 5;						}
const int Editor::getTrackYStart()	{	return from_y;							}
const int Editor::getEditorYStart()	{	return from_y+barHeight+20;				}
const int Editor::getYEnd()			{	return 20+height + from_y+barHeight+20;	} // FIXME - what about to_y ???
const int Editor::getWidth()		{	return width;							}

int Editor::getYScrollInPixels()
{
    std::cout << "Edit base class called. This should not happen." << std::endl;
    exit(0);

}



int Editor::snapMidiTickToGrid(int absolute_x)
{
    
    return (int)( round(
                        (float)absolute_x/
                        (float)(sequence->ticksPerBeat()*4 / graphicalTrack->grid->divider)
                        )*(sequence->ticksPerBeat()*4 / graphicalTrack->grid->divider)
                  );
}

}
